from dataclasses import dataclass
from typing import List

from opperai._client import Client
from opperai.types.datasets import (
    DatasetEntry,
    DatasetEntryResponse,
    DatasetEntryUpdate,
)


@dataclass
class AsyncDataset:
    _client: Client
    _dataset_uuid: str

    async def add(self, entry: DatasetEntry) -> str:
        """Add an entry to the dataset."""
        return await self._client.datasets.add(
            dataset_uuid=self._dataset_uuid, entry=entry
        )

    async def get_entries(
        self, offset: int = 0, limit: int = 100
    ) -> List[DatasetEntryResponse]:
        """Get entries from the dataset."""
        return await self._client.datasets.get_entries(
            dataset_uuid=self._dataset_uuid, offset=offset, limit=limit
        )

    async def get_entry(self, entry_uuid: str) -> DatasetEntryResponse:
        """Get a specific entry from the dataset."""
        return await self._client.datasets.get_entry(
            dataset_uuid=self._dataset_uuid, entry_uuid=entry_uuid
        )

    async def update_entry(
        self, entry_uuid: str, entry: DatasetEntryUpdate
    ) -> DatasetEntryResponse:
        """Update a specific entry in the dataset."""
        return await self._client.datasets.update_entry(
            dataset_uuid=self._dataset_uuid, entry_uuid=entry_uuid, entry=entry
        )

    async def delete_entry(self, entry_uuid: str) -> bool:
        """Delete a specific entry from the dataset."""
        return await self._client.datasets.delete_entry(
            dataset_uuid=self._dataset_uuid, entry_uuid=entry_uuid
        )


class AsyncDatasets:
    _client: Client = None

    def __init__(self, client: Client = None):
        if client is None:
            client = Client()

        self._client = client

    def get(self, dataset_uuid: str) -> AsyncDataset:
        """Get a dataset by its UUID."""
        return AsyncDataset(self._client, dataset_uuid)

    async def add(self, dataset_uuid: str, entry: DatasetEntry) -> str:
        """Add an entry to a dataset."""
        return await self._client.datasets.add(dataset_uuid=dataset_uuid, entry=entry)

    async def get_entries(
        self, dataset_uuid: str, offset: int = 0, limit: int = 100
    ) -> List[DatasetEntryResponse]:
        """Get entries from a dataset."""
        return await self._client.datasets.get_entries(
            dataset_uuid=dataset_uuid, offset=offset, limit=limit
        )

    async def get_entry(
        self, dataset_uuid: str, entry_uuid: str
    ) -> DatasetEntryResponse:
        """Get a specific entry from a dataset."""
        return await self._client.datasets.get_entry(
            dataset_uuid=dataset_uuid, entry_uuid=entry_uuid
        )

    async def update_entry(
        self, dataset_uuid: str, entry_uuid: str, entry: DatasetEntryUpdate
    ) -> DatasetEntryResponse:
        """Update a specific entry in a dataset."""
        return await self._client.datasets.update_entry(
            dataset_uuid=dataset_uuid, entry_uuid=entry_uuid, entry=entry
        )

    async def delete_entry(self, dataset_uuid: str, entry_uuid: str) -> bool:
        """Delete a specific entry from a dataset."""
        return await self._client.datasets.delete_entry(
            dataset_uuid=dataset_uuid, entry_uuid=entry_uuid
        )



import json
from http import HTTPStatus

import httpx
from httpx_sse import aconnect_sse, connect_sse
from opperai import __version__
from opperai.types import Errors
from opperai.types.exceptions import (
    ContentPolicyViolationError,
    OpperTimeoutError,
    RateLimitError,
    RequestValidationError,
    StructuredGenerationError,
)


def _prepare_error(response):
    error = Errors.model_validate(response.json())
    error = error.errors[0]
    status_code = response.status_code

    if status_code == HTTPStatus.BAD_REQUEST:
        if error.type == "StructuredGenerationError":
            raise StructuredGenerationError(error.message, error.detail)
        if error.type == "ContentPolicyViolationError":
            raise ContentPolicyViolationError(error.message, error.detail)
    if status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
        if error.type == "RequestValidationError":
            raise RequestValidationError(error.message, error.detail)
    if status_code == HTTPStatus.TOO_MANY_REQUESTS:
        if error.type == "RateLimitError":
            raise RateLimitError(error.message, error.detail)


class _async_http_client:
    def __init__(self, api_key: str, api_url: str, timeout: float):
        self.session = httpx.AsyncClient(
            base_url=api_url,
            headers={
                "X-OPPER-API-KEY": f"{api_key}",
                "User-Agent": f"opper-python/{__version__}",
            },
            timeout=httpx.Timeout(timeout),
        )

    async def do_request(self, method: str, path: str, **kwargs):
        try:
            response = await self.session.request(method, path, **kwargs)
            if response.status_code >= 400:
                _prepare_error(response)
            return response
        except httpx.TimeoutException as e:
            raise OpperTimeoutError(
                "request timed out",
                "The request to the opper api timed out.",
            ) from e

    async def stream(self, method: str, path: str, **kwargs):
        try:
            async with aconnect_sse(
                self.session,
                method,
                path,
                **kwargs,
            ) as event_source:
                async for sse in event_source.aiter_sse():
                    yield json.loads(sse.data)
        except httpx.TimeoutException as e:
            raise OpperTimeoutError(
                "request timed out",
                "The request to the opper api timed out.",
            ) from e


class _http_client:
    def __init__(self, api_key: str, api_url: str, timeout):
        self.session = httpx.Client(
            base_url=api_url,
            headers={
                "X-OPPER-API-KEY": f"{api_key}",
                "User-Agent": f"opper-python/{__version__}",
            },
            timeout=httpx.Timeout(timeout),
        )

    def do_request(self, method: str, path: str, **kwargs):
        try:
            response = self.session.request(method, path, **kwargs)
            if response.status_code >= 400:
                _prepare_error(response)
            return response
        except httpx.TimeoutException as e:
            raise OpperTimeoutError(
                "request timed out",
                "The request to the opper api timed out.",
            ) from e

    def stream(self, method: str, path: str, **kwargs):
        try:
            with connect_sse(
                self.session,
                method,
                path,
                **kwargs,
            ) as event_source:
                for sse in event_source.iter_sse():
                    yield json.loads(sse.data)
        except httpx.TimeoutException as e:
            raise OpperTimeoutError(
                "request timed out",
                "The request to the opper api timed out.",
            ) from e



import inspect
import itertools
from dataclasses import dataclass
from typing import (
    Any,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
    get_args,
    get_origin,
)

from pydantic import BaseModel, PrivateAttr

from opperai._client import Client
from opperai.core.spans import get_current_span_id
from opperai.core.utils import prepare_examples, prepare_input
from opperai.datasets.datasets import Dataset
from opperai.functions.decorator._schemas import type_to_json_schema
from opperai.types import (
    CallConfiguration,
    CallPayload,
    ChatPayload,
    Example,
    ImageOutput,
    Message,
    StreamingChunk,
)
from opperai.types import Function as FunctionModel
from opperai.types import FunctionResponse as FunctionResponseModel

from ..spans.spans import Span

T = TypeVar("T", bound=Any)


class FunctionResponse(FunctionResponseModel):
    _client: Client = PrivateAttr()

    def __init__(self, client: Client = None, **kwargs):
        super().__init__(**kwargs)
        if not client:
            client = Client()
        self._client = client

    @property
    def span(self):
        return Span(self._client, self.span_id)


class StreamingResponse:
    _client: Client = None
    _head: StreamingChunk = None
    _stream: Iterator[StreamingChunk] = None
    context: List[Dict[str, Any]] = []
    span: Span = None

    def __init__(
        self,
        client: Client = None,
        stream: Iterator[StreamingChunk] = None,
    ):
        if not client:
            client = Client()
        self._client = client

        _stream, _head = itertools.tee(stream)
        self._stream = _stream

        head = next(_head)
        self.span = Span(self._client, head.span_id)
        self.context = head.context if head.context else []

    @property
    def deltas(self) -> Iterator[str]:
        for chunk in self._stream:
            if chunk.delta is not None:
                yield chunk.delta


@dataclass
class Function:
    _client: Client
    _function: FunctionModel

    def chat(
        self,
        messages: List[Message],
        parent_span_id: Optional[str] = None,
        stream: Optional[bool] = False,
    ) -> Union[FunctionResponse, StreamingResponse]:
        if stream:
            if self._function.out_schema is not None:
                raise ValueError(
                    "Cannot stream output for functions structured response"
                )

            response: Iterator[StreamingChunk] = self._client.functions.chat(
                function_path=self._function.path,
                data=ChatPayload(messages=messages, parent_span_uuid=parent_span_id),
                stream=stream,
            )

            return StreamingResponse(client=self._client, stream=response)

        response: FunctionResponseModel = self._client.functions.chat(
            function_path=self._function.path,
            data=ChatPayload(messages=messages, parent_span_uuid=parent_span_id),
            stream=stream,
        )

        return FunctionResponse(client=self._client, **response.model_dump())

    def dataset(self) -> Dataset:
        return Dataset(self._client, self._function.dataset_uuid)

    def delete(self) -> bool:
        return self._client.functions.delete(uuid=self._function.uuid)

    def flush_cache(self) -> bool:
        return self._client.functions.flush_cache(uuid=self._function.uuid)

    def update(self, **kwargs) -> "Function":
        updated = self._function.model_dump(exclude_none=True)
        kwargs["input_schema"] = None
        kwargs["out_schema"] = None
        if "input_type" in kwargs and kwargs["input_type"] is not None:
            kwargs["input_schema"] = type_to_json_schema(kwargs["input_type"])
            del kwargs["input_type"]

        if "output_type" in kwargs and kwargs["output_type"] is not None:
            kwargs["out_schema"] = type_to_json_schema(kwargs["output_type"])
            del kwargs["output_type"]

        for key, value in kwargs.items():
            updated[key] = value

        updated_model = FunctionModel.model_validate(updated)
        updated_function = self._client.functions.update(
            uuid=self._function.uuid, function=updated_model
        )
        self._function = updated_function

        return self

    def call(
        self,
        input: Any = None,
        examples: Optional[List[Example]] = None,
        configuration: Optional[CallConfiguration] = None,
    ):
        payload = CallPayload(
            input=input,
            examples=examples,
        )
        if configuration:
            payload.configuration = configuration

        return self._client.functions.call(
            uuid=self._function.uuid,
            payload=payload,
        )


@dataclass
class Functions:
    _client: Client = None

    def __init__(self, client: Client = None):
        if client is None:
            client = Client()

        self._client = client

    def create(
        self,
        path: str,
        instructions: str,
        description: Optional[str] = None,
        input_type: Optional[Any] = None,
        output_type: Optional[Any] = None,
        model: Optional[str] = None,
    ) -> Function:
        try:
            function = self.get(path=path)
            if function:
                return function.update(
                    instructions=instructions,
                    description=description,
                    input_type=input_type,
                    output_type=output_type,
                    model=model,
                )
        except Exception:
            pass

        if input_type:
            input_schema = type_to_json_schema(input_type)
        if output_type:
            output_schema = type_to_json_schema(output_type)

        function = self._client.functions.create(
            FunctionModel(
                path=path,
                instructions=instructions,
                description=description,
                input_schema=input_schema if input_type else None,
                out_schema=output_schema if output_type else None,
                model=model,
            )
        )

        return Function(self._client, function)

    def get(self, uuid: str = None, path: str = None) -> Optional[Function]:
        if uuid is not None:
            function = self._client.functions.get(uuid=uuid)
        elif path is not None:
            function = self._client.functions.get(path=path)
        else:
            raise ValueError("Either uuid or path must be provided")

        if not function:
            return None

        return Function(self._client, function)

    def delete(self, uuid: str = None, path: str = None) -> bool:
        if uuid is not None:
            return self._client.functions.delete(uuid=uuid)
        elif path is not None:
            return self._client.functions.delete(path=path)
        else:
            raise ValueError("Either uuid or path must be provided")

    def call(
        self,
        name: str = None,
        instructions: str = "you are a helpful assistant",
        input_type: Optional[Any] = None,
        input: Any = None,
        output_type: Optional[type[T]] = None,
        model: Optional[str] = None,
        examples: Optional[List[Example]] = None,
        configuration: Optional[CallConfiguration] = None,
        parent_span_id: Optional[str] = None,
    ) -> Tuple[T, FunctionResponse]:
        """Calls a function
        Arguments:
            name: str: the name of the function, if not provided, it will be generated from the instructions
            instructions: str: the instructions for the function
            input_type: Any: the input type for the function
            input: Any: the input to the function
            output_type: Any: the output type for the function
                There is one special output type:
                    - `ImageOutput`: the output will be an image
            model: str: the model to use for the function
            examples: List[Example]: A list of examples to help guide the function's response.

        Returns:
            tuple[Any, FunctionResponse]: the output of the function and the response object. The type of the output is determined by the output_type. If the output_type is a `Pydantic` model, the output will be validated against the schema.
        """
        if (
            output_type
            and isinstance(output_type, type)
            and issubclass(output_type, ImageOutput)
        ):
            res = self._client.generate_image(prompt=input, configuration=configuration)
            return res

        input_schema = type_to_json_schema(input_type)
        output_schema = type_to_json_schema(output_type)

        _input = prepare_input(input)
        _examples = prepare_examples(examples)

        call_payload = CallPayload(
            name=name,
            instructions=instructions,
            input_schema=input_schema,
            input=_input,
            output_schema=output_schema,
            model=model,
            examples=_examples,
            parent_span_uuid=parent_span_id
            if parent_span_id
            else get_current_span_id(),
        )
        if configuration:
            call_payload.configuration = configuration

        res = self._client.call(call_payload)

        if output_type is not None:
            if inspect.isclass(output_type) and issubclass(output_type, BaseModel):
                return output_type.model_validate(res.json_payload), FunctionResponse(
                    client=self._client, **res.model_dump()
                )
            elif (
                (get_origin(output_type) == list or get_origin(output_type) is List)
                and inspect.isclass(get_args(output_type)[0])
                and issubclass(get_args(output_type)[0], BaseModel)
            ):
                return [
                    get_args(output_type)[0].model_validate(item)
                    for item in res.json_payload
                ], FunctionResponse(client=self._client, **res.model_dump())
            else:
                return res.json_payload, FunctionResponse(
                    client=self._client, **res.model_dump()
                )

        return res.message, FunctionResponse(client=self._client, **res.model_dump())



from typing import List

from opperai.core._http_clients import _async_http_client
from opperai.types.datasets import (
    DatasetEntry,
    DatasetEntryResponse,
    DatasetEntryUpdate,
)
from opperai.types.exceptions import APIError


class AsyncDatasets:
    def __init__(self, http_client: _async_http_client):
        self.http_client = http_client

    async def add(self, dataset_uuid: str, entry: DatasetEntry) -> str:
        response = await self.http_client.do_request(
            "POST",
            f"/v1/datasets/{dataset_uuid}",
            json=entry.model_dump(),
        )
        if response.status_code != 200:
            raise APIError(f"Failed to add entry with status {response.status_code}")
        return response.json()

    async def get_entries(
        self, dataset_uuid: str, offset: int = 0, limit: int = 100
    ) -> List[DatasetEntryResponse]:
        response = await self.http_client.do_request(
            "GET",
            f"/v1/datasets/{dataset_uuid}/entries",
            params={"offset": offset, "limit": limit},
        )
        if response.status_code != 200:
            raise APIError(f"Failed to get entries with status {response.status_code}")
        return [DatasetEntryResponse.model_validate(item) for item in response.json()]

    async def get_entry(
        self, dataset_uuid: str, entry_uuid: str
    ) -> DatasetEntryResponse:
        response = await self.http_client.do_request(
            "GET",
            f"/v1/datasets/{dataset_uuid}/entries/{entry_uuid}",
        )
        if response.status_code != 200:
            raise APIError(f"Failed to get entry with status {response.status_code}")
        return DatasetEntryResponse.model_validate(response.json())

    async def update_entry(
        self, dataset_uuid: str, entry_uuid: str, entry: DatasetEntryUpdate
    ) -> DatasetEntryResponse:
        response = await self.http_client.do_request(
            "PUT",
            f"/v1/datasets/{dataset_uuid}/entries/{entry_uuid}",
            json=entry.model_dump(),
        )
        if response.status_code != 200:
            raise APIError(f"Failed to update entry with status {response.status_code}")
        return DatasetEntryResponse.model_validate(response.json())

    async def delete_entry(self, dataset_uuid: str, entry_uuid: str) -> bool:
        response = await self.http_client.do_request(
            "DELETE",
            f"/v1/datasets/{dataset_uuid}/entries/{entry_uuid}",
        )
        if response.status_code != 200:
            raise APIError(f"Failed to delete entry with status {response.status_code}")
        return response.json()



import asyncio
import contextlib
import inspect
import os
import threading
from functools import wraps
from typing import List, Union, get_args, get_origin, get_type_hints

from opperai._client import AsyncClient, Client
from opperai._opper import AsyncOpper, Opper
from opperai.core.utils import convert_function_call_to_json
from opperai.functions.async_functions import AsyncFunctionResponse
from opperai.functions.functions import Function, FunctionResponse
from opperai.types import (
    CacheConfiguration,
)
from pydantic import BaseModel

from ._schemas import get_output_schema

_thread_local = threading.local()


@contextlib.contextmanager
def span_id_context():
    _thread_local.span_id = None
    try:
        yield
    finally:
        del _thread_local.span_id


def get_last_span_id() -> str:
    """Retrieve the last span ID from thread-local storage."""
    return getattr(_thread_local, "span_id", None)


def fn(
    _func=None,
    *,
    path=None,
    client: Union[Client, AsyncClient] = None,
    json_encoder=None,
    model=None,
    few_shot=None,
    few_shot_count=None,
    cache_config: CacheConfiguration = None,
):
    """Decorator to to create a function in OpperAI's API

    This decorator can automatically handle both synchronous and asynchronous functions,
    wrapping them to perform necessary setup before invoking the OpperAI API. It allows
    for the specification of various parameters to customize the API request, including
    the use of few-shot learning, caching, and custom JSON encoding.

    Parameters:
    - path (str, optional): The API path for the function. Defaults to the decorated function's name.
    - client (Client or AsyncClient, optional): The OpperAI client instance to use. If not provided, a new instance is created.
    - json_encoder (JSONEncoder, optional): Custom JSON encoder for serializing the request payload.
    - model (str, optional): The model to use for the request. If not provided, defaults to the environment's default model.
    - few_shot (bool, optional): Whether to enable few-shot learning. Defaults to False or the environment setting.
    - few_shot_count (int, optional): The number of few-shot examples to use. Only relevant if few_shot is True.
    - cache_config (CacheConfiguration, optional): Configuration for caching API responses.

    Returns:
    - A decorated function that, when called, sets up the necessary OpperAI function and invokes the API, handling both synchronous and asynchronous execution as needed.

    Examples:
    >>> from opperai import fn
    >>> from pydantic import BaseModel

    >>> class Data(BaseModel):
    >>>     keywords: str
    >>>     sentiment: str

    >>> @fn
    >>> def extract(text = str) -> Data:
    >>>     ''' Extract keywords and sentiment from text. '''

    >>> result = extract(text="Opper is a lovely API that wraps the world of AI")
    >>> print(results)
    """

    def decorator(func):
        func_path = path or func.__name__
        setup_done = False
        client: Union[Client, AsyncClient] = None
        opper: Union[Opper, AsyncOpper] = None
        function: Function = None

        def _create_function(opper: Opper):
            return opper.functions.create(
                path=func_path,
                instructions=f"Operation: {func.__name__}\n\nOperation description: {func.__doc__}",
                description=func.__doc__,
                output_type=get_output_schema(func),
                model=model if model else os.environ.get("OPPER_DEFAULT_MODEL", None),
            )

        async def _create_function_async(opper: AsyncOpper):
            return await opper.functions.create(
                path=func_path,
                instructions=f"Operation: {func.__name__}\n\nOperation description: {func.__doc__}",
                description=func.__doc__,
                output_type=get_output_schema(func),
                model=model if model else os.environ.get("OPPER_DEFAULT_MODEL", None),
            )

        def setup():
            nonlocal setup_done, client, function, opper
            if setup_done:
                return

            if isinstance(client, AsyncClient):
                client = Client(api_key=client.api_key, api_url=client.api_url)
                opper = Opper(client)
            elif isinstance(client, Client):
                opper = Opper(client)
            else:
                client = Client()
                opper = Opper(client)

            function = _create_function(opper)

            if asyncio.iscoroutinefunction(func):
                client = AsyncClient(api_key=client.api_key, api_url=client.api_url)
                opper = AsyncOpper(client)
            else:
                client = client

            setup_done = True

        async def async_setup():
            nonlocal setup_done, client, function, opper
            if setup_done:
                return

            if isinstance(client, AsyncClient):
                opper = AsyncOpper(client)
            elif isinstance(client, Client):
                client = AsyncClient(api_key=client.api_key, api_url=client.api_url)
                opper = AsyncOpper(client)
            else:
                client = AsyncClient()
                opper = AsyncOpper(client)

            function = await _create_function_async(opper)

            setup_done = True

        def _unmarshal_response(response, return_type):
            if return_type is not None:
                if inspect.isclass(return_type) and issubclass(return_type, BaseModel):
                    response = return_type(**response)
                elif (
                    (get_origin(return_type) == list or get_origin(return_type) == List)
                    and inspect.isclass(get_args(return_type)[0])
                    and issubclass(get_args(return_type)[0], BaseModel)
                ):
                    response = [get_args(return_type)[0](**item) for item in response]
            return response

        @wraps(func)
        async def async_wrapper(*args, **kwargs):
            answer, _ = await async_call(*args, **kwargs)
            return answer

        async def async_call(*args, **kwargs):
            await async_setup()

            input = convert_function_call_to_json(func, *args, **kwargs)
            _response = await function.call(
                input=input,
            )

            _thread_local.span_id = _response.span_id

            return_type = get_type_hints(func).get("return")
            answer = _unmarshal_response(_response.json_payload, return_type)

            response = AsyncFunctionResponse(client=client, **_response.model_dump())

            return answer, response

        @wraps(func)
        def sync_wrapper(*args, **kwargs):
            answer, _ = sync_call(*args, **kwargs)
            return answer

        def sync_call(*args, **kwargs):
            setup()
            nonlocal function

            input = convert_function_call_to_json(func, *args, **kwargs)
            _response = function.call(
                input=input,
            )

            _thread_local.span_id = _response.span_id

            return_type = get_type_hints(func).get("return")
            answer = _unmarshal_response(_response.json_payload, return_type)

            response = FunctionResponse(client=client, **_response.model_dump())

            return answer, response

        if asyncio.iscoroutinefunction(func):
            async_wrapper.call = async_call
            return async_wrapper
        else:
            sync_wrapper.call = sync_call
            return sync_wrapper

    if _func is None:
        return decorator
    else:
        return decorator(_func)



from typing import List, Optional

from opperai.core._http_clients import _http_client
from opperai.core.spans import get_current_span_id
from opperai.types.exceptions import APIError
from opperai.types.indexes import (
    Document,
    Filter,
    Index,
    RetrievalResponse,
)
from pydantic import BaseModel, Field


class RetrieveQuery(BaseModel):
    q: str = Field(description="Query string")
    k: int = Field(default=3, description="Number of documents to retrieve")
    filters: Optional[List[Filter]] = Field(
        default=None, description="Filters to apply"
    )
    parent_span_uuid: Optional[str] = None


class Indexes:
    def __init__(self, http_client: _http_client):
        self.http_client = http_client

    def create(
        self, name: str, embedding_model: str = "text-embedding-ada-002"
    ) -> Index:
        """Create an index

        This method allows for the creation of an index with the specified name. If the creation is successful, it returns an instance of the Index class, representing the newly created index.

        Args:
            name (str): The name of the index to be created.
            embedding_model (str, optional): The embedding model to use for indexing. Defaults to "text-embedding-ada-002".
        Returns:
            Index: An instance of the Index class, representing the newly created index.

        Raises:
            APIError: If the index creation fails due to an API error.

        Examples:
            >>> from opperai import Client
            >>> client = Client(api_key="your_api_key_here")
            >>> index = client.indexes.create(name="my_new_index")
            >>> print(index)
            Index(id='123', name='my_new_index')

        """
        response = self.http_client.do_request(
            "POST",
            "/v1/indexes",
            json={
                "name": name,
                "embedding_model": embedding_model,
            },
        )
        if response.status_code != 200:
            raise APIError(f"Failed to create index with status {response.status_code}")
        return Index.model_validate(response.json())

    def delete(self, uuid: str) -> bool:
        """Delete an index

        This method allows for the deletion of an index by specifying its unique identifier. If the deletion is successful, the method returns True. If the index cannot be found or the deletion fails, the method returns False.

        Args:
            uuid (str): The unique identifier of the index to delete.

        Returns:
            bool: True if the index was successfully deleted, False otherwise.

        Raises:
            APIError: If the index deletion fails due to an API error.

        Examples:
            >>> from opperai import Client
            >>> client = Client(api_key="your_api_key_here")
            >>> response = client.indexes.delete(uuid="123")
            >>> print(response)
            True

        """
        response = self.http_client.do_request(
            "DELETE",
            f"/v1/indexes/{uuid}",
        )
        if response.status_code == 404:
            return False
        if response.status_code != 200:
            raise APIError(f"Failed to delete index with status {response.status_code}")
        return True

    def get(self, uuid: str = None, name: str = None) -> Optional[Index]:
        """Retrieve an index

        This method fetches an index either by its unique identifier or by its name. If the index is found, it returns an instance of the Index class representing the index's details. If no index matches the given ID or name, or if both parameters are omitted, None is returned.

        Args:
            uuid (str, optional): The unique identifier of the index to retrieve. Defaults to None.
            name (str, optional): The name of the index to retrieve. Defaults to None.

        Returns:
            Optional[Index]: An instance of the Index class if the index is found, otherwise None.

        Raises:
            ValueError: If both `uuid` and `name` are provided, indicating ambiguous parameters.
            APIError: If the retrieval fails due to an API error.

        Examples:
            >>> from opperai import Client
            >>> client = Client(api_key="your_api_key_here")
            >>> index_by_uuid = client.indexes.get(uuid="123")
            >>> print(index_by_uuid)
            Index(id='123', name='my_index', ...)

            >>> index_by_name = client.indexes.get(name="my_index")
            >>> print(index_by_name)
            Index(id='123', name='my_index', ...)

        Note:
            It is recommended to provide either `id` or `name`, but not both, to avoid ambiguity. If neither is provided, the method will return None.
        """
        if uuid is None and name is None:
            raise ValueError("Either uuid or name must be provided")
        if uuid is not None and name is not None:
            raise ValueError("Only one of id or name should be provided")

        if uuid is not None:
            return self._get_by_uuid(uuid)
        if name is not None:
            return self._get_by_name(name)

    def _get_by_uuid(self, uuid: str) -> Optional[Index]:
        response = self.http_client.do_request(
            "GET",
            f"/v1/indexes/{uuid}",
        )
        if response.status_code == 404:
            return None
        if response.status_code != 200:
            raise APIError(f"Failed to get index with status {response.status_code}")

        return Index.model_validate(response.json())

    def _get_by_name(self, name: str) -> Optional[Index]:
        indexes = self.list()
        for index in indexes:
            if index.name == name:
                return index
        return None

    def list(self) -> List[Index]:
        """List all indexes

        This method retrieves a list of all indexes available in the system. It returns a list of Index instances, each representing the details of an index.

        Returns:
            List[Index]: A list of Index instances representing all the indexes available.

        Raises:
            APIError: If the listing of indexes fails due to an API error.

        Examples:
            >>> from opperai import Client
            >>> client = Client(api_key="your_api_key_here")
            >>> indexes = client.indexes.list()
            >>> for index in indexes:
            ...     print(index)
            Index(id='123', name='my_index', ...)
            Index(id='124', name='another_index', ...)

        """
        response = self.http_client.do_request(
            "GET",
            "/v1/indexes",
        )
        if response.status_code != 200:
            raise APIError(f"Failed to list indexes with status {response.status_code}")
        return [Index.model_validate(item) for item in response.json()]

    def upload_file(self, uuid: str, file_path: str, **kwargs):
        """Upload a file to an index

        This method uploads a file to a specified index by its unique identifier. The file is uploaded to a pre-signed URL obtained from the server, and then the file is registered with the index.

        Args:
            id (int): The unique identifier of the index to which the file is being uploaded.
            file_path (str): The path to the file that needs to be uploaded.
            **kwargs: Additional keyword arguments that will be included in the registration request.

        Returns:
            dict: A dictionary containing the server's response to the file registration.

        Raises:
            APIError: If the file upload or registration fails due to an API error.

        Examples:
            >>> from opperai import Client
            >>> client = Client(api_key="your_api_key_here")
            >>> response = client.indexes.upload_file(id=123, file_path="/path/to/file.txt")
            >>> print(response)
            {'message': 'File uploaded and registered successfully'}

        """
        # Get upload URL
        upload_url_response = self.http_client.do_request(
            "GET",
            f"/v1/indexes/{uuid}/upload_url",
            params={"filename": file_path.split("/")[-1]},
        )
        if upload_url_response.status_code != 200:
            raise APIError(
                f"Failed to get upload URL with status {upload_url_response.status_code}"
            )
        upload_url_data = upload_url_response.json()

        # Upload file
        with open(file_path, "rb") as file:
            files = {"file": (file_path.split("/")[-1], file)}
            upload_response = self.http_client.do_request(
                "POST",
                upload_url_data["url"],
                files=files,
                data=upload_url_data["fields"],
            )
            if upload_response.status_code not in [200, 204]:
                raise APIError(
                    f"Failed to upload file with status {upload_response.status_code}"
                )

        # Register file
        register_file_response = self.http_client.do_request(
            "POST",
            f"/v1/indexes/{uuid}/register_file",
            json={"uuid": upload_url_data["uuid"], **kwargs},
        )
        if register_file_response.status_code != 200:
            raise APIError(
                f"Failed to register file with status {register_file_response.status_code}"
            )

        return register_file_response.json()

    def index(self, uuid: str, doc: Document) -> Document:
        """Index a document

        This method adds a document to a specified index by its unique identifier. The document is indexed within the system, making it searchable or retrievable according to the index's configuration.

        Args:
            id (int): The unique identifier of the index where the document will be indexed.
            doc (Document): The document to be indexed.

        Returns:
            Document: An instance of the Document class, representing the indexed document with any server-assigned metadata.

        Raises:
            APIError: If the document indexing fails due to an API error.

        Examples:
            >>> from opperai import Client
            >>> from opperai.types.indexes import Document
            >>> client = Client(api_key="your_api_key_here")
            >>> doc = Document(content="Example content")
            >>> indexed_doc = client.indexes.index(id=123, doc=doc)
            >>> print(indexed_doc)
            Document(id='456', content='Example content', ...)

        """
        response = self.http_client.do_request(
            "POST",
            f"/v1/indexes/{uuid}/index",
            json=doc.model_dump(),
        )
        if response.status_code != 200:
            raise APIError(f"Failed to add document with status {response.status_code}")
        return Document.model_validate(response.json())

    def retrieve(
        self,
        uuid: str,
        query: str,
        k: int,
        filters: Optional[List[Filter]] = None,
        parent_span_uuid: Optional[str] = None,
        **kwargs,
    ) -> List[RetrievalResponse]:
        """Retrieve documents from an index

        This method retrieves documents from a specified index based on a query. It supports filtering and limiting the number of responses.

        Args:
            id (int): The unique identifier of the index from which documents are to be retrieved.
            query (str): The query string used to search and retrieve documents.
            k (int): The maximum number of documents to retrieve.
            filters (Optional[List[Filter]], optional): A list of filters to apply to the retrieval process.

        Returns:
            List[RetrievalResponse]: A list of RetrievalResponse instances, each representing a document that matches the query criteria.

        Raises:
            APIError: If the retrieval process fails due to an API error.

        Examples:
            >>> from opperai import Client
            >>> from opperai.types.indexes import Filter
            >>> client = Client(api_key="your_api_key_here")
            >>> filters = [Filter(field="date", value="2021-01-01", operation="gte")]
            >>> responses = client.indexes.retrieve(id=123, query="example query", k=5, filters=filters)
            >>> for response in responses:
            ...     print(response)
            RetrievalResponse(id='doc123', score=0.98, ...)

        """
        response = self.http_client.do_request(
            "POST",
            f"/v1/indexes/{uuid}/query",
            json={
                **RetrieveQuery(
                    q=query,
                    k=k,
                    filters=filters,
                    parent_span_uuid=parent_span_uuid or get_current_span_id(),
                ).model_dump(),
                **kwargs,
            },
        )
        if response.status_code != 200:
            raise APIError(
                f"Failed to retrieve index {id} with status {response.status_code}"
            )
        return [RetrievalResponse.model_validate(item) for item in response.json()]



from __future__ import annotations

from contextlib import contextmanager
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from typing import Any, Dict, Iterator, List

from opperai._client import Client
from opperai.core.spans import _current_span_id
from opperai.types.spans import GenerationIn, GenerationOut, SpanMetric
from opperai.types.spans import Span as SpanModel


@dataclass
class Span:
    _client: Client
    uuid: str

    def update(self, **kwargs) -> Span:
        self._client.spans.update(span_uuid=self.uuid, **kwargs)
        return self

    def delete(self) -> bool:
        return self._client.spans.delete(span_uuid=self.uuid)

    def save_example(self, **kwargs) -> str:
        return self._client.spans.save_example(uuid=self.uuid, **kwargs)

    def save_metric(self, dimension: str, value: float, comment: str = None) -> bool:
        return self._client.spans.save_metric(
            uuid=self.uuid,
            metric=SpanMetric(
                dimension=dimension,
                value=value,
                comment=comment,
            ),
        )

    def save_generation(
        self,
        duration_ms: int,
        called_at: datetime = None,
        input: str = None,
        response: str = None,
        model: str = None,
        messages: List[Dict[str, Any]] = None,
        prompt_tokens: int = None,
        completion_tokens: int = None,
        total_tokens: int = None,
        error: str = None,
        cost: float = None,
    ) -> GenerationOut:
        # if called_at is None, use the current time minus the duration_ms
        if called_at is None:
            now = datetime.now(timezone.utc)
            called_at = now - timedelta(milliseconds=duration_ms)

        return self._client.spans.save_generation(
            uuid=self.uuid,
            generation=GenerationIn(
                called_at=called_at,
                duration_ms=duration_ms,
                input=input,
                response=response,
                model=model,
                messages=messages,
                prompt_tokens=prompt_tokens,
                completion_tokens=completion_tokens,
                total_tokens=total_tokens,
                error=error,
                cost=cost,
            ),
        )


class Spans:
    _client: Client = None

    def __init__(self, client: Client = None):
        if client is None:
            client = Client()

        self._client = client

    @contextmanager
    def start(
        self,
        name: str,
        input: str = None,
        meta: dict = None,
        parent_span_id: str = None,
    ) -> Iterator[Span]:
        span = self.start_span(name, input, meta, parent_span_id)
        try:
            yield span
        finally:
            span.end()

    def start_span(
        self,
        name: str,
        input: str = None,
        meta: dict = None,
        parent_span_id: str = None,
    ) -> Span:
        span, token = self._create_span(name, input, meta, parent_span_id)

        def end_span():
            if not hasattr(span, "_ended"):
                span.update(end_time=datetime.now(timezone.utc))
                _current_span_id.reset(token)
                span._ended = True

        span.end = end_span

        return span

    def _create_span(self, name, input, meta, parent_span_id):
        parent_span_id = parent_span_id if parent_span_id else _current_span_id.get()
        span_model = self._client.spans.create(
            SpanModel(
                name=name,
                input=input,
                start_time=datetime.now(timezone.utc),
                parent_uuid=parent_span_id,
                meta=meta,
            )
        )
        span = Span(self._client, span_model.uuid)
        token = _current_span_id.set(str(span.uuid))

        return span, token

    @property
    def current_span(self) -> Span:
        return Span(self._client, _current_span_id.get())

    def get_span(self, span_id: str) -> Span:
        return Span(self._client, span_id)



import json
from http import HTTPStatus
from typing import Any, Dict
from uuid import UUID

from opperai.core._http_clients import _http_client
from opperai.core.utils import DateTimeEncoder
from opperai.types.exceptions import APIError
from opperai.types.spans import GenerationIn, GenerationOut, Span, SpanMetric


class Spans:
    def __init__(self, http_client: _http_client):
        self.http_client = http_client

    def create(self, span: Span, **kwargs) -> Span:
        """Create a new span in the system.

        This method sends a POST request to the server to create a new span with the provided details. The span's data is serialized into JSON format, excluding any unset attributes, before being sent as the request payload.

        Args:
            span (Span): An instance of the Span class, representing the span to be created.
            **kwargs: Additional keyword arguments that will be included in the request payload.

        Returns:
            str: The UUID of the newly created span as a string.

        Raises:
            APIError: If the span creation fails due to an API error.

        Examples:
            >>> from opperai import Client
            >>> from opperai.types.spans import Span
            >>> client = Client(api_key="your_api_key_here")
            >>> new_span = Span(
            ...     name="New Span",
            ...     start_time="2021-07-21T17:32:28Z",
            ...     end_time="2021-07-21T18:32:28Z",
            ...     metadata={"key": "value"}
            ... )
            >>> span_uuid = client.spans.create(new_span)
            >>> print(span_uuid)
            '123e4567-e89b-12d3-a456-426614174000'
        """
        span_data = span.model_dump(exclude_none=True)
        json_payload = json.dumps(span_data, cls=DateTimeEncoder)
        response = self.http_client.do_request(
            "POST",
            "/v1/spans",
            content=json_payload,
        )
        if response.status_code != 200:
            raise APIError(
                f"Failed to create span {span.name} with status {response.status_code}"
            )
        return Span.model_validate(response.json())

    def update(self, span_uuid: UUID, **kwargs) -> Span:
        """Update an existing span in the system

        This method sends a PUT request to the server to update an existing span identified by its UUID with the provided details. The span's data is serialized into JSON format, excluding any unset attributes, before being sent as the request payload.

        Args:
            span_uuid (UUID): The UUID of the span to be updated.
            **kwargs: Additional keyword arguments that will be included in the request payload.

        Returns:
            str: The UUID of the updated span as a string.

        Raises:
            APIError: If the span update fails due to an API error.

        Examples:
            >>> from opperai import Client
            >>> from opperai.types.spans import Span
            >>> client = Client(api_key="your_api_key_here")
            >>> updated_span = Span(
            ...     name="Updated Span",
            ...     start_time="2021-07-21T17:32:28Z",
            ...     end_time="2021-07-21T18:32:28Z",
            ...     metadata={"key": "new_value"}
            ... )
            >>> span_uuid = "123e4567-e89b-12d3-a456-426614174000"
            >>> updated_uuid = client.spans.update(span_uuid, updated_span)
            >>> print(updated_uuid)
            '123e4567-e89b-12d3-a456-426614174000'
        """
        span = Span(uuid=span_uuid, **kwargs)
        json_payload = json.dumps(
            span.model_dump(exclude_none=True), cls=DateTimeEncoder
        )
        response = self.http_client.do_request(
            "PUT",
            f"/v1/spans/{span.uuid}",
            content=json_payload,
        )
        if response.status_code != 200:
            raise APIError(
                f"Failed to update span `{span.name}` with status {response.status_code}"
            )

        return Span.model_validate(response.json())

    def delete(self, span_uuid: UUID) -> bool:
        """Delete an existing span from the system

        This method sends a DELETE request to the server to remove a span identified by its UUID. If the deletion is successful, the method returns True, indicating that the span has been successfully removed from the system.

        Args:
            span_uuid (UUID): The UUID of the span to be deleted.

        Returns:
            bool: True if the span was successfully deleted, False otherwise.

        Raises:
            APIError: If the span deletion fails due to an API error.

        Examples:
            >>> from opperai import Client
            >>> client = Client(api_key="your_api_key_here")
            >>> span_uuid = "123e4567-e89b-12d3-a456-426614174000"
            >>> result = client.spans.delete(span_uuid)
            >>> print(result)
            True
        """
        response = self.http_client.do_request(
            "DELETE",
            f"/v1/spans/{span_uuid}",
        )
        if response.status_code != HTTPStatus.NO_CONTENT:
            raise APIError(
                f"Failed to delete span `{span_uuid}` with status {response.status_code}"
            )

        return True

    def save_example(self, uuid: str, **kwargs) -> str:
        """Saves a span as an example in the system

        This method sends a POST request to the server's `/v1/spans/{uuid}/save_examples` endpoint. The UUID in the URL is replaced with the UUID of the span for which the example is being saved.

        Args:
            uuid (str): The UUID of the span for which the example is being saved.
            **kwargs: Additional keyword arguments that can be used for future extensions or to include additional data in the request. These are not used in the current implementation.

        Returns:
            str: The UUID of the span for which the example was saved.

        Raises:
            APIError: If the server responds with a status code other than 200, indicating that the example was not successfully saved for the span.

        Examples:
            >>> from opperai import Client
            >>> client = Client(api_key="your_api_key_here")
            >>> span_uuid = "123e4567-e89b-12d3-a456-426614174000"
            >>> result_uuid = client.spans.save_example(span_uuid)
            >>> print(result_uuid)
            '123e4567-e89b-12d3-a456-426614174000'
        """
        response = self.http_client.do_request(
            "POST",
            f"/v1/spans/{uuid}/save_examples",
        )
        if response.status_code != 200:
            raise APIError(
                f"Failed to save examples for span {uuid} with status {response.status_code}"
            )

        return response.json()["uuid"]

    def save_metric(self, uuid: str, metric: SpanMetric, **kwargs) -> Dict[str, Any]:
        """Saves a metric for a specific span

        This method sends a POST request to the server's `/v1/spans/{uuid}/metrics` endpoint, including the metric data for the span identified by the given UUID. The metric data is serialized into JSON format, excluding any unset attributes, before being sent as the request payload.

        Args:
            uuid (str): The UUID of the span for which the metric is being saved.
            metric (SpanMetric): The metric object containing the data for the span.
            **kwargs: Additional keyword arguments that can be used for future extensions or to include additional data in the request. These are not used in the current implementation.

        Returns:
            Dict[str, Any]: A dictionary containing the server's response. The structure of this dictionary is determined by the server's response schema for the metric submission endpoint.

        Raises:
            APIError: If the server responds with a status code other than 200, indicating that the metric was not successfully saved for the span.

        Examples:
            >>> from opperai import Client
            >>> from opperai.types.spans import SpanMetric
            >>> client = Client(api_key="your_api_key_here")
            >>> metric = SpanMetric(name="Load Time", value=123)
            >>> span_uuid = "123e4567-e89b-12d3-a456-426614174000"
            >>> result = client.spans.save_metric(span_uuid, metric)
            >>> print(result)
        """
        response = self.http_client.do_request(
            "POST",
            f"/v1/spans/{uuid}/metrics",
            json=metric.model_dump(exclude_unset=True),
        )
        if response.status_code != 200:
            raise APIError(
                f"Failed to add metric for span {uuid} with status {response.status_code}"
            )

        return response.json()

    def save_generation(
        self,
        uuid: str,
        generation: GenerationIn,
    ) -> GenerationOut:
        response = self.http_client.do_request(
            "POST",
            f"/v1/spans/{uuid}/generation",
            content=json.dumps(
                generation.model_dump(exclude_none=True), cls=DateTimeEncoder
            ),
        )

        if response.status_code != 200:
            raise APIError(
                f"Failed to save generation for span {uuid} with status {response.status_code}"
            )

        return GenerationOut.model_validate(response.json())



from typing import Optional


class APIError(Exception):
    message: str

    def __init__(self, message: str):
        super().__init__(message)
        self.message = message


class OpperBaseException(Exception):
    """Base class for all exceptions raised by the opperai API."""

    type: Optional[str] = None
    message: Optional[str] = None
    detail: Optional[str] = None

    def __init__(
        self,
        type: Optional[str] = None,
        message: Optional[str] = None,
        detail: Optional[str] = None,
    ):
        super().__init__(message)
        self.type = type if type else self.type
        self.message = message if message else self.message
        self.detail = detail if detail else self.detail


class OpperTimeoutError(OpperBaseException):
    """Raised when the request to the opper api times out."""

    type: str = "OpperTimeoutError"


class RateLimitError(OpperBaseException):
    """Raised when the rate limit is exceeded."""

    type: str = "RateLimitError"


class StructuredGenerationError(OpperBaseException):
    """Raised when the structured generation fails."""

    type: str = "StructuredGenerationError"


class ContentPolicyViolationError(OpperBaseException):
    """Raised when the content policy is violated."""

    type: str = "ContentPolicyViolationError"


class RequestValidationError(OpperBaseException):
    """Raised when the request validation fails."""

    type: str = "RequestValidationError"



import datetime
from typing import Any, Dict, List, Optional, Union

from pydantic import UUID4, BaseModel, Field


class RetrievalResponse(BaseModel):
    content: str
    metadata: Dict[str, Any]


class DocumentIn(BaseModel):
    key: Optional[str] = None
    content: Optional[str] = Field(None, min_length=1)
    metadata: Optional[Dict[str, Any]] = None


class Document(DocumentIn):
    uuid: UUID4


class Index(BaseModel):
    uuid: UUID4
    name: str
    created_at: datetime.datetime


class Filter(BaseModel):
    key: str
    operation: str = Field(
        ..., description="The operation to perform on the key", pattern=r"^=|!=|>|<|in$"
    )
    value: Union[str, int, float, List[Union[str, int, float]]]



from datetime import datetime
from typing import Any, Dict, List, Optional
from uuid import UUID

from pydantic import BaseModel


class GenerationIn(BaseModel):
    called_at: datetime
    duration_ms: int
    model: Optional[str] = None
    input: Optional[str] = None
    response: Optional[str] = None
    prompt_tokens: Optional[int] = None
    completion_tokens: Optional[int] = None
    total_tokens: Optional[int] = None
    error: Optional[str] = None
    messages: Optional[List[Dict[str, Any]]] = None
    cost: Optional[float] = None


class GenerationOut(GenerationIn):
    uuid: UUID


class Span(BaseModel):
    uuid: Optional[UUID] = None
    name: Optional[str] = None
    input: Optional[str] = None
    output: Optional[str] = None
    start_time: Optional[datetime] = None
    parent_uuid: Optional[UUID] = None
    end_time: Optional[datetime] = None
    error: Optional[str] = None
    meta: Optional[dict] = None
    evaluations: Optional[dict] = None
    score: Optional[int] = None
    generation: Optional[GenerationIn] = None


class SpanMetric(BaseModel):
    dimension: Optional[str] = None
    value: Optional[float] = None
    comment: Optional[str] = None



import inspect
from dataclasses import dataclass
from typing import (
    Any,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
    get_args,
    get_origin,
)

from pydantic import BaseModel, PrivateAttr

from opperai._client import AsyncClient
from opperai.core.spans import get_current_span_id
from opperai.datasets.async_datasets import AsyncDataset
from opperai.functions.decorator._schemas import type_to_json_schema
from opperai.types import (
    CallConfiguration,
    CallPayload,
    ChatPayload,
    Example,
    ImageOutput,
    Message,
    StreamingChunk,
)
from opperai.types import Function as FunctionModel
from opperai.types import FunctionResponse as FunctionResponseModel

from ..spans.async_spans import AsyncSpan
from .functions import T, prepare_examples, prepare_input


class AsyncFunctionResponse(FunctionResponseModel):
    _client: AsyncClient = PrivateAttr()

    def __init__(self, client: AsyncClient = None, **kwargs):
        super().__init__(**kwargs)
        if not client:
            client = AsyncClient()
        self._client = client

    @property
    def span(self):
        return AsyncSpan(self._client, self.span_id)


class AsyncStreamingResponse:
    _client: AsyncClient = None
    stream: Iterator[StreamingChunk] = None
    _context: Optional[List[Dict[str, Any]]] = None
    _span_id: Optional[str] = None

    def __init__(
        self, client: AsyncClient = None, stream: Iterator[StreamingChunk] = None
    ):
        if not client:
            client = AsyncClient()
        self._client = client
        self.stream = stream

    async def initialize(self):
        self.stream = await self.stream
        try:
            first_chunk = await self.stream.__anext__()
            self._span_id = (
                first_chunk.span_id if first_chunk.span_id is not None else None
            )
            self._context = (
                first_chunk.context if first_chunk.context is not None else []
            )
        except StopAsyncIteration:
            self._span_id = None
            self._context = []

    @property
    async def deltas(self):
        async for chunk in self.stream:
            if chunk.delta is not None:
                yield chunk.delta

    @property
    def span(self) -> AsyncSpan:
        return AsyncSpan(self._client, self._span_id)


@dataclass
class AsyncFunction:
    _client: AsyncClient
    _function: FunctionModel

    async def chat(
        self,
        messages: List[Message],
        parent_span_id: Optional[str] = None,
        stream: Optional[bool] = False,
    ) -> Union[AsyncFunctionResponse, AsyncStreamingResponse]:
        if stream:
            if self._function.out_schema is not None:
                raise ValueError(
                    "Cannot stream output for functions structured response"
                )

            response = self._client.functions.chat(
                function_path=self._function.path,
                data=ChatPayload(messages=messages, parent_span_uuid=parent_span_id),
                stream=stream,
            )

            res = AsyncStreamingResponse(client=self._client, stream=response)
            await res.initialize()
            return res

        response: FunctionResponseModel = await self._client.functions.chat(
            function_path=self._function.path,
            data=ChatPayload(messages=messages, parent_span_uuid=parent_span_id),
            stream=stream,
        )

        return AsyncFunctionResponse(client=self._client, **response.model_dump())

    async def call(
        self,
        input: Any = None,
        examples: Optional[List[Example]] = None,
        configuration: Optional[CallConfiguration] = None,
    ) -> Tuple[T, AsyncFunctionResponse]:
        payload = CallPayload(
            input=input,
            examples=examples,
        )
        if configuration:
            payload.configuration = configuration

        return await self._client.functions.call(
            uuid=self._function.uuid,
            payload=payload,
        )

    async def delete(self) -> bool:
        return await self._client.functions.delete(uuid=self._function.uuid)

    async def flush_cache(self) -> bool:
        return await self._client.functions.flush_cache(uuid=self._function.uuid)

    async def update(self, **kwargs) -> "AsyncFunction":
        updated = self._function.model_dump(exclude_none=True)
        kwargs["input_schema"] = None
        kwargs["out_schema"] = None
        if "input_type" in kwargs and kwargs["input_type"] is not None:
            kwargs["input_schema"] = type_to_json_schema(kwargs["input_type"])
            del kwargs["input_type"]
        if "output_type" in kwargs and kwargs["output_type"] is not None:
            kwargs["out_schema"] = type_to_json_schema(kwargs["output_type"])
            del kwargs["output_type"]

        for key, value in kwargs.items():
            updated[key] = value

        updated_model = FunctionModel.model_validate(updated)
        updated_function = await self._client.functions.update(
            uuid=self._function.uuid, function=updated_model
        )
        self._function = updated_function

        return self

    def dataset(self) -> AsyncDataset:
        return AsyncDataset(self._client, self._function.dataset_uuid)


@dataclass
class AsyncFunctions:
    _client: AsyncClient = None

    def __init__(self, client: AsyncClient = None):
        if client is None:
            client = AsyncClient()

        self._client = client

    async def create(
        self,
        path: str,
        instructions: str,
        description: Optional[str] = None,
        input_type: Optional[Any] = None,
        output_type: Optional[Any] = None,
        model: Optional[str] = None,
    ) -> AsyncFunction:
        try:
            function = await self.get(path=path)
            if function:
                return await function.update(
                    instructions=instructions,
                    description=description,
                    input_type=input_type,
                    output_type=output_type,
                    model=model,
                )
        except Exception:
            pass

        if input_type:
            input_schema = type_to_json_schema(input_type)
        if output_type:
            output_schema = type_to_json_schema(output_type)

        function = await self._client.functions.create(
            FunctionModel(
                path=path,
                instructions=instructions,
                description=description,
                input_schema=input_schema if input_type else None,
                out_schema=output_schema if output_type else None,
                model=model,
            )
        )

        return AsyncFunction(self._client, function)

    async def get(self, uuid: str = None, path: str = None) -> Optional[AsyncFunction]:
        if uuid is not None:
            function = await self._client.functions.get(uuid=uuid)
        elif path is not None:
            function = await self._client.functions.get(path=path)
        else:
            raise ValueError("Either id or name must be provided")

        if not function:
            return None

        return AsyncFunction(self._client, function)

    async def delete(self, uuid: str = None, path: str = None) -> bool:
        if uuid is not None:
            return await self._client.functions.delete(uuid=uuid)
        elif path is not None:
            return await self._client.functions.delete(path=path)
        else:
            raise ValueError("Either id or name must be provided")

    async def call(
        self,
        name: str = None,
        instructions: str = "you are a helpful assistant",
        input_type: Optional[Any] = None,
        input: Any = str,
        output_type: Optional[type[T]] = None,
        model: Optional[str] = None,
        examples: Optional[List[Example]] = None,
        configuration: Optional[CallConfiguration] = None,
        parent_span_id: Optional[str] = None,
    ) -> Tuple[T, AsyncFunctionResponse]:
        """Calls a function
        Arguments:
            name: str: the name of the function, if not provided, it will be generated from the instructions
            instructions: str: the instructions for the function
            input_type: Any: the input type for the function
            input: Any: the input to the function
            output_type: Any: the output type for the function
                There is one special output type:
                    - `ImageOutput`: the output will be an image
            model: str: the model to use for the function
            examples: List[Example]: A list of examples to help guide the function's response

        Returns:
            tuple[Any, FunctionResponse]: the output of the function and the response object. The type of the output is determined by the output_type. If the output_type is a `Pydantic` model, the output will be validated against the schema.
        """
        if (
            output_type
            and isinstance(output_type, type)
            and issubclass(output_type, ImageOutput)
        ):
            res = await self._client.generate_image(prompt=input)
            return res

        input_schema = type_to_json_schema(input_type)
        output_schema = type_to_json_schema(output_type)

        _input = prepare_input(input)
        _examples = prepare_examples(examples)

        call_payload = CallPayload(
            name=name,
            instructions=instructions,
            input_schema=input_schema,
            input=_input,
            output_schema=output_schema,
            model=model,
            examples=_examples,
            parent_span_uuid=parent_span_id
            if parent_span_id
            else get_current_span_id(),
        )
        if configuration:
            call_payload.configuration = configuration

        res: FunctionResponseModel = await self._client.call(call_payload)

        if output_type is not None:
            if inspect.isclass(output_type) and issubclass(output_type, BaseModel):
                return output_type.model_validate(
                    res.json_payload
                ), AsyncFunctionResponse(client=self._client, **res.model_dump())
            elif (
                (get_origin(output_type) == list or get_origin(output_type) is List)
                and inspect.isclass(get_args(output_type)[0])
                and issubclass(get_args(output_type)[0], BaseModel)
            ):
                return [
                    get_args(output_type)[0].model_validate(item)
                    for item in res.json_payload
                ], AsyncFunctionResponse(client=self._client, **res.model_dump())
            else:
                return res.json_payload, AsyncFunctionResponse(
                    client=self._client, **res.model_dump()
                )

        return res.message, AsyncFunctionResponse(
            client=self._client, **res.model_dump()
        )



from opperai.functions.async_functions import AsyncFunctions
from opperai.functions.functions import Functions
from opperai.indexes.async_indexes import AsyncIndexes
from opperai.indexes.indexes import Indexes
from opperai.spans.async_spans import AsyncSpans
from opperai.spans.spans import Spans

from ._client import AsyncClient, Client

DEFAULT_API_URL = "https://api.opper.ai"
DEFAULT_TIMEOUT = 120


class Opper:
    def __init__(self, client: Client = None, api_key: str = None):
        if client is not None:
            if not isinstance(client, Client):
                raise ValueError("Client must be an instance of Client")
        if api_key is not None:
            client = Client(api_key=api_key)
        if client is None:
            client = Client()

        self.client: Client = client
        self.functions: Functions = Functions(client)
        self.indexes: Indexes = Indexes(client)
        self.spans: Spans = Spans(client)  # deprecated
        self.traces: Spans = self.spans
        self.call = self.functions.call


class AsyncOpper(Opper):
    def __init__(self, client: AsyncClient = None, api_key: str = None):
        if client is not None:
            if not isinstance(client, AsyncClient):
                raise ValueError("Client must be an instance of AsyncClient")
        if api_key is not None:
            client = AsyncClient(api_key=api_key)
        if client is None:
            client = AsyncClient()

        self.client: AsyncClient = client
        self.functions: AsyncFunctions = AsyncFunctions(client)
        self.indexes: AsyncIndexes = AsyncIndexes(client)
        self.spans: AsyncSpans = AsyncSpans(client)  # deprecated
        self.traces: AsyncSpans = self.spans
        self.call = self.functions.call
